from openapi_core.validation.response.datatypes import OpenAPIResponse


class ClientOpenAPIResponseFactory(object):
    @classmethod
    def create(cls, response):
        oapi_response = OpenAPIResponse(
            data=response.content, status_code=response.status_code, mimetype=response.headers.get("content-type"),
        )
        oapi_response.payload = response.json()
        return oapi_response



from collections import namedtuple
from functools import wraps
from importlib import import_module
from inspect import iscoroutinefunction
from pathlib import Path
from types import ModuleType
from typing import Callable, Union, Optional
from urllib.parse import urlsplit

from openapi_core import create_spec
from openapi_core.schema.specs.models import Spec
from openapi_core.shortcuts import RequestValidator, ResponseValidator
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import JSONResponse, Response
from stringcase import snakecase

from pyotr.utils import get_spec_from_file
from pyotr.validation.requests import StarletteOpenAPIRequest
from pyotr.validation.responses import StarletteOpenAPIResponse

Operation = namedtuple("Operation", "path method")


class Application(Starlette):
    def __init__(
        self,
        spec: Union[Spec, dict],
        *,
        module: Optional[Union[str, ModuleType]] = None,
        validate_responses: bool = True,
        enforce_case: bool = True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if not isinstance(spec, Spec):
            spec = create_spec(spec)
        self.spec = spec
        self.validate_responses = validate_responses
        self.enforce_case = enforce_case
        self._server_paths = {urlsplit(server.url).path for server in self.spec.servers}

        self._operations = {
            oper.operation_id: Operation(path, method)
            for path, path_spec in spec.paths.items()
            for method, oper in path_spec.operations.items()
        }

        if module is not None:
            if isinstance(module, str):
                module = _load_module(module)

            for operation_id, operation in self._operations.items():
                name = operation_id
                if "." in name:
                    base, name = name.rsplit(".", 1)
                    base_module = _load_module(f"{module.__name__}.{base}")
                else:
                    base_module = module
                if self.enforce_case:
                    name = snakecase(name)
                try:
                    endpoint_fn = getattr(base_module, name)
                except AttributeError as e:
                    raise RuntimeError(f"The function `{base_module}.{name}` does not exist!") from e
                self.set_endpoint(endpoint_fn, operation_id=operation_id)

    def set_endpoint(self, endpoint_fn: Callable, *, operation_id: Optional[str] = None):
        """ Sets endpoint function for a given `operationId`.

            If the `operation_id` is not given, it will try to determine it based on the function name.
        """
        if operation_id is None:
            operation_id = endpoint_fn.__name__
        if self.enforce_case and operation_id not in self._operations:
            operation_id_key = {snakecase(op_id): op_id for op_id in self._operations}.get(operation_id)
        else:
            operation_id_key = operation_id
        try:
            operation = self._operations[operation_id_key]
        except KeyError as e:
            raise ValueError(f"Unknown operationId: {operation_id}.") from e

        @wraps(endpoint_fn)
        async def wrapper(request: Request, **kwargs) -> Response:
            openapi_request = await StarletteOpenAPIRequest(request)
            request_validation = RequestValidator(self.spec).validate(openapi_request)
            request_validation.raise_for_errors()

            if iscoroutinefunction(endpoint_fn):
                response = await endpoint_fn(request, **kwargs)
            else:
                response = endpoint_fn(request, **kwargs)
            if isinstance(response, dict):
                response = JSONResponse(response)

            # TODO: pass a list of operation IDs to specify which responses not to validate
            if self.validate_responses:
                ResponseValidator(self.spec).validate(
                    openapi_request, StarletteOpenAPIResponse(response)
                ).raise_for_errors()
            return response

        for server_path in self._server_paths:
            self.add_route(server_path + operation.path, wrapper, [operation.method])

    def endpoint(self, operation_id: Union[Callable, str]):
        """ Decorator for setting endpoints.

            If used without arguments, it will try to determine the `operationId` based on the
            decorated function name:

                @app.endpoint
                def foo_bar(request):
                    # sets the endpoint for operationId fooBar

            Otherwise, the `operationId` can be set explicitly:

                @app.endpoint('fooBar'):
                def my_endpoint():
                    ...
        """
        if callable(operation_id):
            self.set_endpoint(operation_id)
            return operation_id
        else:

            def decorator(fn):
                self.set_endpoint(fn, operation_id=operation_id)
                return fn

            return decorator

    @classmethod
    def from_file(cls, path: Union[Path, str], *args, **kwargs) -> "Application":
        """ Creates an instance of the class by loading the spec from a local file.
        """
        spec = get_spec_from_file(path)
        return cls(spec, *args, **kwargs)


def _load_module(name: str) -> ModuleType:
    """ Helper function to load module based on its dotted-string name.
    """
    try:
        module = import_module(name)
    except ModuleNotFoundError as e:
        raise RuntimeError(f"The module `{name}` does not exist!") from e
    else:
        return module



from string import Formatter
from urllib.parse import parse_qs, urlsplit, urljoin, urlencode, urlunsplit

from openapi_core.schema.operations.models import Operation
from openapi_core.validation.request.datatypes import RequestParameters, OpenAPIRequest


class ClientOpenAPIRequest(OpenAPIRequest):
    def __init__(self, host_url: str, op_spec: Operation):
        self.spec = op_spec
        self._url_parts = urlsplit(host_url)

        formatter = Formatter()
        self.url_vars = [var for _, var, _, _ in formatter.parse(op_spec.path_name) if var is not None]
        self._path_pattern = self._url_parts.path + op_spec.path_name

        self.full_url_pattern = urljoin(host_url, self._path_pattern)
        self.method = op_spec.http_method.lower()
        self.body = None
        self.parameters = RequestParameters(path={}, query=parse_qs(self._url_parts.query), header={}, cookie={},)
        self.mimetype = NotImplemented

    @property
    def url(self):
        url_parts = self._url_parts._asdict()
        url_parts["path"] = self._path_pattern.format(**self.parameters.path)
        url_parts["query"] = urlencode(self.parameters.query)
        return urlunsplit(url_parts.values())

    def prepare(self, *args, **kwargs):
        len_vars = len(self.url_vars)
        if len(args) != len_vars:
            error_message = f"Incorrect arguments: {self.spec.operation_id} accepts"
            if len_vars:
                error_message += (
                    f" {len_vars} positional argument{'s' if len_vars > 1 else ''}:" f" {', '.join(self.url_vars)}"
                )
            else:
                error_message += f" no positional arguments"
            raise RuntimeError(error_message)
        self.parameters.header = kwargs.pop("headers_", None) or {}
        self.parameters.path = dict(zip(self.url_vars, args))
        self.parameters.query = kwargs
        self.mimetype = self.parameters.header.get("content-type", None)
        return self

    @property
    def headers(self):
        return self.parameters.header



from .client import ClientOpenAPIResponseFactory
from .starlette import StarletteOpenAPIResponseFactory

ClientOpenAPIResponse = ClientOpenAPIResponseFactory.create
StarletteOpenAPIResponse = StarletteOpenAPIResponseFactory.create



from openapi_core.validation.response.datatypes import OpenAPIResponse


class StarletteOpenAPIResponseFactory:
    @classmethod
    def create(cls, response):
        return OpenAPIResponse(
            data=response.body, status_code=response.status_code, mimetype=response.headers.get("content-type"),
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pyotr',
 'pyotr.validation',
 'pyotr.validation.requests',
 'pyotr.validation.responses']

package_data = \
{'': ['*']}

install_requires = \
['http3>=0.6.3,<0.7.0',
 'httpx>=0.7.2,<0.8.0',
 'openapi-core>=0.13.1,<0.14.0',
 'pyyaml>=5.1,<6.0',
 'starlette>=0.13,<0.14',
 'stringcase>=1.2,<2.0',
 'typing-extensions>=3.7,<4.0',
 'werkzeug>=1.0.0,<2.0.0']

extras_require = \
{':extra == "uvicorn"': ['uvicorn>=0.9.0,<0.10.0']}

setup_kwargs = {
    'name': 'pyotr',
    'version': '0.4.3',
    'description': 'Python OpenAPI-to-REST (and back) framework ',
    'long_description': 'Pyotr\n=====\n\n[![Documentation Status](https://readthedocs.org/projects/pyotr/badge/?version=latest)](https://pyotr.readthedocs.io/en/latest/?badge=latest)\n[![CI builds](https://b11c.semaphoreci.com/badges/pyotr.svg?style=shields)](https://b11c.semaphoreci.com/badges/pyotr.svg?style=shields)\n\n**Pyotr** is a Python library for serving and consuming REST APIs based on \n[OpenAPI](https://swagger.io/resources/open-api/) specifications. Its name is acronym of "Python OpenAPI to REST".\n\nThe project consists of two separate libraries that can be used independently:\n\n* `pyotr.server` is a [Starlette](https://www.starlette.io)-based framework for serving OpenAPI-based services. \n  It is functionally very similar to [connexion](https://connexion.readthedocs.io), except that it aims to be fully \n  [ASGI](https://asgi.readthedocs.io)-compliant. \n* `pyotr.client` is a HTTP client for consuming OpenAPI-based services.\n\n**WARNING:** This is still very much work in progress and not quite ready for production usage. Until version 1.0 is \nreleased, any version can be expected to break backward compatibility.\n\n\nQuick Start\n-----------\n\n### Server\n\n    from pyotr.server import Application\n    from some.path import endpoints\n    \n    app = Application.from_file("path/to/openapi.yaml", module=endpoints)\n    \n### Client\n\n    from pyotr.client import Client\n    \n    client = Client.from_file("path/to/openapi.yaml")\n    result = client.some_endpoint_id("path", "variables", "query_var"="example")\n',
    'author': 'Berislav Lopac',
    'author_email': 'berislav@lopac.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pyotr.readthedocs.io',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
